/*
 * <OWNER> = Alvaro Arcas Garcia
 * <YEAR> = 2014
 *
 * In the original BSD license, the occurrence of "copyright holder" in the 3rd clause read "ORGANIZATION", placeholder for "University of California". In the original BSD license, both occurrences of the phrase "COPYRIGHT HOLDERS AND CONTRIBUTORS" in the disclaimer read "REGENTS AND CONTRIBUTORS".
 *
 * Here is the license template:
 *
 * Copyright (c) <2014>, <Alvaro Arcas Garcia>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

library ArcNeuralNetwork.DataSet;
import 'package:collection/equality.dart';


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////           CLASS DATASET                          //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
Class DataSet: represent a dataSet, with instance for train and test.
      Parameters:
        -  title => title of the data --> String
        -  _instances => all the instance of the dataSet -->  List <Instance>
        -  _labels => the name of attributes -->  List <String>
        -  isSupervised => true if the dataSet contains class values, false if not --> bool
        -  numValues => number of attributes --> int
        -  numClassValues => number of class values --> int
      Constructors:
        - Supervised => create a supervised dataSet
        - NoSupervised => create a no supervised dataSet
        - List => create a supervised/no supervised dataSet with a list of values
        - File => create a supervised/no supervised dataSet with the content of a file
      Methods:
        - Add Instance
        - Remove Instance
        - Add Row
        - Remove Row
        - Get labels
        - Get Instances
        - Get Row
        - Get train
        - Get test
        - Set labels
        - Set Instances
        - Set Row
*/

class DataSet {

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      PARAMETERS        ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  String title;
  List <Instance> _instances;
  List <String> _labels;
  bool _isSupervised;
  int _numValues;
  int _numClassValues;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      CONSTRUCTORS      ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


  DataSet.Supervised(String title, int numValues, int numClassValues) {

    if(title == null){
      this.title = "";
    }else{
      this.title = title;
    }

    if(numValues == null)
      throw ("Number of values can not be null!");
    if(numValues <=0)
      throw ("Number of values has to be greatter than 0!");
    this._numValues = numValues;

    if(numClassValues == null)
      throw ("Number of class values can not be null!");
    if(numClassValues <=0)
      throw ("Number of class values has to be greatter than 0!");
    this._numClassValues = numClassValues;

    this._instances = [];
    this._labels = [];
    this._isSupervised = true;

  }

  DataSet.NoSupervised(String title, int numValues){

    if(title == null){
      this.title = "";
    }else{
      this.title = title;
    }

    if(numValues == null)
      throw ("Number of values can not be null!");
    if(numValues <=0)
      throw ("Number of values has to be greatter than 0!");
    this._numValues = numValues;

    this._instances = [];
    this._labels = [];
    this._isSupervised = false;
  }

  DataSet.List(String title, int numValues,  List<String> labels, List<List<double>> values, [int numClassValues]){

    if(numClassValues != null){
      if(title == null){
        this.title = "";
      }else{
        this.title = title;
      }

      if(numValues == null)
        throw ("Number of values can not be null!");
      if(numValues <=0)
        throw ("Number of values has to be greatter than 0!");
      this._numValues = numValues;

      if(numClassValues == null)
        throw ("Number of class values can not be null!");
      if(numClassValues <=0)
        throw ("Number of class values has to be greatter than 0!");
      this._numClassValues = numClassValues;

      this._instances = [];
      this._labels = [];
      this._isSupervised = true;
    }else{
      if(title == null){
        this.title = "";
      }else{
        this.title = title;
      }

      if(numValues == null)
        throw ("Number of values can not be null!");
      if(numValues <=0)
        throw ("Number of values has to be greatter than 0!");
      this._numValues = numValues;

      this._instances = [];
      this._labels = [];
      this._isSupervised = false;
    }
    this.instances = values;
    this.labels = labels;
  }

  DataSet.File(String content){
    List<String> lines = content.split("\n");
    this.title = lines[0];
    List <String> parametersTemp = lines[1].split(";");
    if(parametersTemp.length == 1 || parametersTemp.length==2){
      if(parametersTemp.length==1){
        if(int.parse(parametersTemp[0])==0)
          throw("Number of values has to be greatter than 0");
        this._numValues = int.parse(parametersTemp[0]);
      }else{
        if(int.parse(parametersTemp[0])==0)
          throw("Number of values has to be greatter than 0");
        if(int.parse(parametersTemp[1])==0)
          throw("Number of class values has to be greatter than 0");
        this._numValues = int.parse(parametersTemp[0]);
        this._numClassValues = int.parse(parametersTemp[1]);
      }
    }else{
      throw("You set "+ parametersTemp.length.toString()+" parameters instead of 2: values and class values(only for supervised dataSet)");
    }

    if(numClassValues!= null){
      this._isSupervised = true;
    }else{
      this._isSupervised = false;
    }

    this.labels = lines[2].split(";");
    this._instances = [];

    for(int i = 3; i < lines.length; i++){
      List<String> temp = lines[i].split(";");
      List<double> temp2 = [];
      for(String value in temp){
        temp2.add(double.parse(value));
      }
      this.addInstance(temp2);
    }

  }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      GETTERS AND SETTERS  /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  * NAME: labels
  * RETURNS: a list of strings with the labels values.
  * PARAMETERS: nothing
  */

  get labels => this._labels;

  /*
  * NAME: labels
  * RETURNS: nothing
  * PARAMETERS:
  * * List <String> labels-- A list with the values of the labels to be set.
  */

  set labels(List <String> labels){
    if(_isSupervised){
      if(labels.length != (this._numValues + this._numClassValues))
        throw ("You set a list of "+labels.length.toString()+" labels but the dataSet has "+this._labels.length.toString()+" labels");
      this._labels = labels;
    }else{
      if(labels.length != this._numValues)
        throw ("You set a list of "+labels.length.toString()+" labels but the dataSet has "+this._labels.length.toString()+" labels");
      this._labels = labels;
    }
  }

  /*
  * NAME: setLabelAt
  * RETURNS: nothing
  * PARAMETERS:
  * * int index -- index of the label to set
  * * String label -- new name for label
  */

  setLabelAt(int index, String label){
    this._labels[index] = label;
  }

  /*
  * NAME: isSupervised
  * RETURNS: bool
  * PARAMETERS: nothing
  */

  get isSupervised => this._isSupervised;

  /*
  * NAME: isSupervised
  * RETURNS: nothing
  * PARAMETERS:
  * * bool isSupervised -- bool to change the dataSet to supervised or no supervised.
  * * * When you change to a supervised dataSet the las row will be the class value row.
  * * * If you change to a no supervised dataSet all row will be no class values
  */

  set isSupervised(bool isSupervised){
    if(this._isSupervised == true && isSupervised == false){
      for(Instance tempInstance in this._instances){
        tempInstance.values.addAll(tempInstance.classValues);
        tempInstance.numValues+=tempInstance.numClassValues;
        tempInstance.numClassValues = null;
        tempInstance.classValues = null;
      }
      this._numValues+=this._numClassValues;
      this._numClassValues = null;
      this._isSupervised = false;

    }
    if(this._isSupervised == false && isSupervised == true){
      for(Instance tempInstance in this._instances){
        tempInstance.classValues = [];
        tempInstance.numValues--;
        tempInstance.numClassValues = 1;
        tempInstance.classValues.add(tempInstance.values.last);
        tempInstance.values.removeLast();
      }
      this._numValues--;
      this._numClassValues = 1;
      this._isSupervised = true;
    }

  }

  /*
  * NAME: numValues
  * RETURNS: int with the number of values.
  * PARAMETERS: nothing
  */

  get numValues => _numValues;

  /*
  * NAME: numClassValues
  * RETURNS: int with the number of class values.
  * PARAMETERS: nothing
  */

  get numClassValues => _numClassValues;

  /*
  * NAME: instances
  * RETURNS: a list of instances.
  * PARAMETERS: nothing
  */

  get instances => this._instances;

  /*
  * NAME: instances
  * RETURNS: nothing
  * PARAMETERS:
  * * List <List <double>> values -- bidimensional list with the values of all the instances that want to be set.
  */

  set instances (List<List<double>> values){
    this._instances=[];
    for(List<double> value in values){
        this.addInstance(value);
    }
  }

  /*
  * NAME: setRow
  * RETURNS: nothing.
  * PARAMETERS:
  * * int index -- index of the row to be set.
  * * List <double> values -- values to be set.
  * * [optional] String label -- new name for the label of the row
  */

  void setRow(int index, List <double> values, [String label]){
    if(label!= null) {
      this.setLabelAt(index, label);
    }
    if(_isSupervised){
      if(index < this._numValues){
        if(values.length != this._instances.length)
          throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
        int count = 0;
        for(Instance temp in this._instances){
          temp.values[index] = values[count];
          count++;
        }
      }else{
        if(values.length != this._instances.length)
          throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
        index-=this._numValues;
        int count = 0;
        for(Instance temp in this._instances){
          temp.classValues[index] = values[count];
          count++;
        }
      }
    }else{
      if(values.length != this._instances.length)
        throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
      int count = 0;
      for(Instance temp in this._instances){
        temp.values[index] = values[count];
        count++;
      }
    }
  }

  /*
  * NAME: getRow
  * RETURNS: List <double> -- A list with the values of the row
  * PARAMETERS:
  * * int index -- index of the row.
  */

  List<double> getRow(int index){
    List <double> values = [];
    if(index < this._numValues){
      for(Instance temp in this._instances){
        values.add(temp.values[index]);
      }
    }else{
      index -= this._numValues;
      for(Instance temp in this._instances){
        values.add(temp.classValues[index]);
      }
    }
    return values;
  }

  /*
  * NAME: trainSet
  * RETURNS: list of instances with all the instance for train.
  * PARAMETERS:nothing
  */

  List <Instance> get trainSet{
    List <Instance> temp = [];
    for (Instance instance in this.instances){
      if(instance.isForTrain == null){
       throw ("Some instances has not been set for train or test. Before you use the train set or test set you must set the instances for train and test");
      }else{
        if(instance.isForTrain){
          temp.add(instance);
        }
      }
    }
    return temp;
  }

  /*
  * NAME: testSet
  * RETURNS: list of instances with all the instance for test.
  * PARAMETERS:nothing
  */

  List <Instance> get testSet{
    List <Instance> temp = [];
    for (Instance instance in this.instances){
      if(instance.isForTrain == null){
        throw ("Some instances has not been set for train or test. Before you use the train set or test set you must set the instances for train and test");
      }else {
        if (!instance.isForTrain) {
          temp.add(instance);
        }
      }
    }
    return temp;
  }

  /*
  * NAME: trainTestSet
  * RETURNS: nothing
  * PARAMETERS:
  * * int numberInstanceTrain -- number of instance that will be for train. The rest of the instance will be for test
  */

  void set trainTestSet(int numberInstanceTrain){
    if(this._instances.isEmpty){
      throw("The dataSet has no instances to set");
    }else{
      for(int i = 0; i < this.instances.length; i++){
        if(i < numberInstanceTrain){
          this.instances[i].isForTrain = true;
        }else{
          this.instances[i].isForTrain = false;
        }
      }
    }

  }

  /*
  * NAME: instancesValues
  * RETURNS: List<List <double>> -- values of the instances
  * PARAMETERS:nothing
  */

  List<List<double>> get instancesValues{
    List<List<double>> temp = [];
    for(Instance instance in this.instances){
      temp.add(instance.instanceValues);
    }
    return temp;
  }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      METHODS              /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  * NAME: addInstance
  * RETURNS: nothing
  * PARAMETERS:
  * * List <double> values -- values of the new instance
  * * [Optional] bool isForTrain -- if true the instance will use in the trainSet, if false instance will use in the testSet
  */

  void addInstance(List <double> values, [bool isForTrain]){
    if(this.isSupervised){
      if(values.length != (this._numValues + this._numClassValues))
        throw ("You add a instance of "+values.length.toString()+" values, but an instance in this dataSet has "+this._numValues.toString()+" attributes and "+this._numClassValues.toString()+" class values. So an instance must have a total of "+(this._numValues+this._numClassValues).toString()+" values");
      Instance temp = new Instance.Supervised(this._numValues, this._numClassValues, values.getRange(0,this._numValues).toList(),values.getRange(this._numValues, this._numValues + this._numClassValues).toList());
      if(isForTrain != null){
        temp.isForTrain = isForTrain;
      }
      if(!this.hasInstance(temp.instanceValues)){
        this._instances.add(temp);
      }
    }else{
      if(values.length != this._numValues)
        throw ("You add a instance of "+values.length.toString()+" values, but an instance in this dataSet has "+this._numValues.toString()+" attributes. So an instance must have a total of "+this._numValues.toString()+" values");
      Instance temp = new Instance.NoSupervised(this._numValues, values);
      if(isForTrain != null){
        temp.isForTrain = isForTrain;
      }
      if(!this.hasInstance(temp.instanceValues)){
        this._instances.add(temp);
      }
    }
  }

  bool hasInstance(List <double> instance){
    bool repeated = false;
    for(List <double> temp in this.instancesValues){
      Function eq = const ListEquality().equals;
      if(eq(temp, instance)){
       return true;
      }
    }
    return false;
  }

  /*
  * NAME: removeInstance
  * RETURNS: nothing
  * PARAMETERS:
  * * int index -- index of the instance that will be remove
  */

  void removeInstance(int index){
    this._instances.removeAt(index);
  }

  /*
  * NAME: addRow
  * RETURNS: nothing
  * PARAMETERS:
  * * List <double> values -- values of the new row. The length of the list must be the same as the length of the list instances
  * * String label -- name of the label for the new row
  * * bool isClassValue -- if true the new row is a class value, if false is a normal attribute
  */

  void addRow(List <double> values, String label, bool isClassValue){
    if(!_isSupervised){
      if(isClassValue){
        if(values.length != this._instances.length)
          throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
        this._numClassValues = 1;
        int count = 0;
        for(Instance temp in this._instances){
          temp.classValues = [];
          temp.numClassValues = 0;
          temp.addClassValue(values[count]);
          count++;
        }

        this._labels.add(label);
        this._isSupervised = true;

      }else{
        if(values.length != this._instances.length)
          throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
        this._numValues++;
        int count = 0;
        for(Instance temp in this._instances){
          temp.addValue(values[count]);
          count++;
        }
        this._labels.add(label);
      }
    }else{
      if(isClassValue){
        if(values.length != this._instances.length)
          throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
        this._numClassValues++;
        int count = 0;
        for(Instance temp in this._instances){
          temp.addClassValue(values[count]);
          count++;
        }

        this._labels.add(label);
      }else{
        if(values.length != this._instances.length)
          throw ("You set a row of "+values.length.toString()+" values, but the dataSet has rows of "+this._instances.length.toString()+" values.");
        this._numValues++;
        int count = 0;
        for(Instance temp in this._instances){
          temp.addValue(values[count]);
          count++;
        }
        this._labels.insert(_numValues, label);
      }
    }

  }

  /*
  * NAME: removeRow
  * RETURNS: nothing
  * PARAMETERS:
  * * int index -- index of the row to delete.
  */

  void removeRow(int index){
    this._labels.removeAt(index);
    if(_isSupervised){
      if(index < this._numValues){
        this._numValues --;
        for(Instance temp in this._instances){
          temp.numValues--;
          temp.values.removeAt(index);
        }
      }else{
        this._numClassValues--;
        index-=this._numValues;
        for(Instance temp in this._instances){
          temp.numClassValues--;
          temp.classValues.removeAt(index);
        }
        if(this._numClassValues == 0){
          this.isSupervised = false;
        }
      }
    }else{
      this._numValues--;
      for(Instance temp in this._instances){
        temp.numValues--;
        temp.values.removeAt(index);
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////           CLASS INSTANCE                         //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
Class Instance: represents a instance of a dataSet
      Parameters:
        - values => double values of the attributes --> List <double>
        - classValues => double values of the class values --> List <double>
        - isForTrain => if true, the instance is for train --> bool
        - numValues => number of attributes --> int
        - numClassValues => number of class values --> int
      Constructors:
        - Instance => default empty instance
        - Supervised => create a supervised instance
        - NoSupervised => create a no supervised instance
      Methods:
        - Add Value
        - Remove Value
        - Add classValue
        - Remove classValue
        - Get instanceValues
        - Set instanceValues
*/

class Instance{

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      PARAMETERS        ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  List <double> values;
  List <double> classValues;
  bool isForTrain;
  int numValues;
  int numClassValues;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      CONSTRUCTORS      ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  Instance(int numValues){

    if(numValues == null)
      throw ("Number of values can not be null!");
    if(numValues <=0)
      throw ("Number of values has to be greatter than 0!");

    this.numValues = numValues;
    this.values = [];

  }

  Instance.Supervised(int numValues, int numClassValues, List <double> values, List <double> classValues){

    if(numValues == null)
      throw ("Number of values can not be null!");
    if(numValues <=0)
      throw ("Number of values has to be greatter than 0!");
    this.numValues = numValues;

    if(numClassValues == null)
      throw ("Number of classValues can not be null!");
    if(numClassValues <=0)
      throw ("Number of classValues has to be greatter than 0!");
    this.numClassValues = numClassValues;

    if(values == null){
      this.values = [];
    }else{
      if(values.length != numValues) {
        throw ("The values list size is different than the number of values!!");
      }else{
        this.values = values;
      }
    }
    if(classValues == null){
      this.classValues = [];
    }else{
      if(classValues.length != numClassValues) {
        throw ("The classValues list size is different than the number of class values!!");
      }else{
        this.classValues = classValues;
      }
    }

  }
  Instance.NoSupervised(int numValues, List <double> values){

    if(numValues == null)
      throw ("Number of values can not be null!");
    if(numValues <=0)
      throw ("Number of values has to be greatter than 0");
    this.numValues = numValues;

    if(values == null){
      this.values = [];
    }else{
      if(values.length != numValues) {
        throw ("The values list size is different than the number of values!!");
      }else{
        this.values = values;
      }
    }

  }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      GETTERS AND SETTERS //////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  * NAME: instanceValues
  * RETURNS: A list of all the values (values + classValues)
  * PARAMETERS:nothing
  */

  get instanceValues{

    List <double> instanceValues = [];
    if(this.values != null){
      instanceValues.addAll(this.values);
    }
    if(this.classValues != null){
      instanceValues.addAll(this.classValues);
    }
    return instanceValues;

  }

  /*
  * NAME: instanceValues
  * RETURNS: nothing
  * PARAMETERS:
  * * List <double> values -- values of the instance.
  */

  set instanceValues(List<double> values){

    if(numClassValues != null){
      if(values.length != (this.numValues + this.numClassValues))
        throw ("The values list size is different than the number of values and class values");
      this.values = values.getRange(0, numValues-1);
      this.classValues = values.getRange(numValues, numClassValues-1);
    }else{
      if(values.length != this.numValues)
        throw ("The values list size is different than the number of values and class values");
      this.values = values;
    }

  }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      METHODS           ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  * NAME: addClassValue
  * RETURNS: nothing
  * PARAMETERS:
  * * double value -- the new classValue to add
  */

  addClassValue(double value){
    this.numClassValues ++;
    this.classValues.add(value);
  }

  /*
  * NAME: removeClassValue
  * RETURNS: nothing
  * PARAMETERS:
  * * int index -- the index of the classValue to remove
  */

  removeClassValue(int index){
    this.numClassValues --;
    this.classValues.removeAt(index);

  }

  /*
  * NAME: addValue
  * RETURNS: nothing
  * PARAMETERS:
  * * double value -- the new value to add
  */

  addValue(double value){
    this.numValues++;
    this.values.add(value);
  }

  /*
  * NAME: removeValue
  * RETURNS: nothing
  * PARAMETERS:
  * * int index -- the index of the value to remove
  */

  removeValue(int index){
    this.numValues--;
    this.values.removeAt(index);
  }
}
