/*
 * <OWNER> = Alvaro Arcas Garcia
 * <YEAR> = 2014
 *
 * In the original BSD license, the occurrence of "copyright holder" in the 3rd clause read "ORGANIZATION", placeholder for "University of California". In the original BSD license, both occurrences of the phrase "COPYRIGHT HOLDERS AND CONTRIBUTORS" in the disclaimer read "REGENTS AND CONTRIBUTORS".
 *
 * Here is the license template:
 *
 * Copyright (c) <2014>, <Alvaro Arcas Garcia>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import "dataset.dart";
import 'package:unittest/unittest.dart';

main(){

  ////////////////////////////////TEST CONSTRUCTORS/////////////////////////////////////////////////////////////////////

  test('SupervisedDataSet', (){
    DataSet dataSetTest = new DataSet.Supervised("TEST",2,2);
    expect(dataSetTest.isSupervised, true);
    expect(dataSetTest.title == "TEST", true);
    expect(dataSetTest.numValues == 2, true);
    expect(dataSetTest.numClassValues == 2, true);
    expect(dataSetTest.instances.isEmpty , true);
    expect(dataSetTest.labels.isEmpty, true);
  });

  test('NoSupervisedDataSet', (){
    DataSet dataSetTest = new DataSet.NoSupervised("TEST",2);
    expect(dataSetTest.isSupervised, false);
    expect(dataSetTest.title == "TEST", true);
    expect(dataSetTest.numValues == 2, true);
    expect(dataSetTest.numClassValues == null, true);
    expect(dataSetTest.instances.isEmpty , true);
    expect(dataSetTest.labels.isEmpty, true);
  });

  test('ListSupervisedDataSet', (){
    List <List <double>> values = [[2.1, 4.1, 2.3],[2.3, 4.3, 2.5],[2.2, 4.2, 2.4]];
    List <String> labels = ["At1","At2","At3"];

    DataSet dataSetTest = new DataSet.List("TEST",2,labels, values, 1);

    expect(dataSetTest.title == "TEST", true);
    expect(dataSetTest.isSupervised, true);
    expect(dataSetTest.numValues == 2, true);
    expect(dataSetTest.numClassValues == 1, true);
    expect(dataSetTest.instances.length == 3 , true);
    expect(dataSetTest.labels.length == 3, true);
    expect(dataSetTest.labels, equals(["At1","At2","At3"]));
    expect(dataSetTest.instances[0].instanceValues, equals([2.1, 4.1, 2.3]));
    expect(dataSetTest.instances[1].instanceValues, equals([2.3, 4.3, 2.5]));
    expect(dataSetTest.instances[2].instanceValues, equals([2.2, 4.2, 2.4]));
    expect(dataSetTest.instances[0].values, equals([2.1, 4.1]));
    expect(dataSetTest.instances[0].classValues, equals([2.3]));
  });

  test('ListNoSupervisedDataSet', (){
    List <List <double>> values = [[2.1, 4.1, 2.3],[2.3, 4.3, 2.5],[2.2, 4.2, 2.4]];
    List <String> labels = ["At1","At2","At3"];

    DataSet dataSetTest = new DataSet.List("TEST",3,labels, values);

    expect(dataSetTest.title == "TEST", true);
    expect(dataSetTest.isSupervised, false);
    expect(dataSetTest.numValues == 3, true);
    expect(dataSetTest.numClassValues == null, true);
    expect(dataSetTest.instances.length == 3 , true);
    expect(dataSetTest.labels.length == 3, true);
    expect(dataSetTest.labels, equals(["At1","At2","At3"]));
    expect(dataSetTest.instances[0].instanceValues, equals([2.1, 4.1, 2.3]));
    expect(dataSetTest.instances[1].instanceValues, equals([2.3, 4.3, 2.5]));
    expect(dataSetTest.instances[2].instanceValues, equals([2.2, 4.2, 2.4]));
    expect(dataSetTest.instances[0].values, equals([2.1, 4.1,2.3]));
    expect(dataSetTest.instances[0].classValues == null, true);
  });

  test('FileSupervisedDataSet',(){
    String dataSetContent =
    "Este es el titulo\n"+
    "5;4\n"+
    "at1;at2:c;at3;at4;at5;at6;at7;at8;c1\n"+
    "0.908675799;0;0;0.62460063;0;0.941860465;0.047666837;0;0.128441513\n"+
    "0.291552505;0;0.590704615;0.586261947;0.189440988;0.660174349;0.410436492;0.035714286;0.249034515\n"+
    "0;0.425709523;0;0.560702865;0;0.25;0.873055747;0.074175824;0.186246435\n"+
    "0.420776228;0.558987194;0;0.182907305;0.347826076;0.591860395;0.526091381;0.016483516;0.444375245\n"+
    "0.202511409;0;0.626686648;0.321885016;0.242236017;0.840116279;0.526843986;0.035714286;0.233337479\n"+
    "0.633561644;0.420701167;0;0.256389711;0.493788805;0.968895491;0.02759659;0.005494505;0.32727047\n"+
    "0.634703196;0.264329442;0;0.848242808;0;0.380813953;0;1;0.515385599\n"+
    "0.342009139;0;0.49925037;0.194089485;0.385093156;0.595930233;0.767185164;0.005494505;0.232839174\n"+
    "0.502283105;0;0;0.648562292;0;0.502906977;0.516808862;0.491758242;0.339603846\n"+
    "0.337899543;0;0;0.480830658;0;0.869186047;0.56698448;0.074175824;0.230970493\n"+
    "0.341095877;0;0.591204395;0.532747593;0.180124218;0.661046581;0.410687461;0.151098901;0.431543564";

    DataSet dataSetTest = new DataSet.File(dataSetContent);

    expect(dataSetTest.title == "Este es el titulo",true);
    expect(dataSetTest.numValues == 5, true);
    expect(dataSetTest.numClassValues == 4, true);
    expect(dataSetTest.instances.length == 11, true);
    expect(dataSetTest.instances[0].instanceValues, equals([0.908675799,0,0,0.62460063,0,0.941860465,0.047666837,0,0.128441513]));
    expect(dataSetTest.instances[10].instanceValues, equals([0.341095877,0,0.591204395,0.532747593,0.180124218,0.661046581,0.410687461,0.151098901,0.431543564]));
    expect(dataSetTest.labels, equals(["at1","at2:c","at3","at4","at5","at6","at7","at8","c1"]));
    expect(dataSetTest.isSupervised == true, true);

  });

  test('FileNoSupervisedDataSet',(){
    String dataSetContent =
    "Este es el titulo\n"+
    "9\n"+
    "at1;at2:c;at3;at4;at5;at6;at7;at8;c1\n"+
    "0.908675799;0;0;0.62460063;0;0.941860465;0.047666837;0;0.128441513\n"+
    "0.291552505;0;0.590704615;0.586261947;0.189440988;0.660174349;0.410436492;0.035714286;0.249034515\n"+
    "0;0.425709523;0;0.560702865;0;0.25;0.873055747;0.074175824;0.186246435\n"+
    "0.420776228;0.558987194;0;0.182907305;0.347826076;0.591860395;0.526091381;0.016483516;0.444375245\n"+
    "0.202511409;0;0.626686648;0.321885016;0.242236017;0.840116279;0.526843986;0.035714286;0.233337479\n"+
    "0.633561644;0.420701167;0;0.256389711;0.493788805;0.968895491;0.02759659;0.005494505;0.32727047\n"+
    "0.634703196;0.264329442;0;0.848242808;0;0.380813953;0;1;0.515385599\n"+
    "0.342009139;0;0.49925037;0.194089485;0.385093156;0.595930233;0.767185164;0.005494505;0.232839174\n"+
    "0.502283105;0;0;0.648562292;0;0.502906977;0.516808862;0.491758242;0.339603846\n"+
    "0.337899543;0;0;0.480830658;0;0.869186047;0.56698448;0.074175824;0.230970493\n"+
    "0.341095877;0;0.591204395;0.532747593;0.180124218;0.661046581;0.410687461;0.151098901;0.431543564";

    DataSet dataSetTest = new DataSet.File(dataSetContent);

    expect(dataSetTest.title == "Este es el titulo",true);
    expect(dataSetTest.numValues == 9, true);
    expect(dataSetTest.numClassValues == null, true);
    expect(dataSetTest.instances.length == 11, true);
    expect(dataSetTest.instances[0].instanceValues, equals([0.908675799,0,0,0.62460063,0,0.941860465,0.047666837,0,0.128441513]));
    expect(dataSetTest.instances[10].instanceValues, equals([0.341095877,0,0.591204395,0.532747593,0.180124218,0.661046581,0.410687461,0.151098901,0.431543564]));
    expect(dataSetTest.labels, equals(["at1","at2:c","at3","at4","at5","at6","at7","at8","c1"]));
    expect(dataSetTest.isSupervised == false, true);

  });

  ////////////////////////////////TEST GETTERS AND SETTERS//////////////////////////////////////////////////////////////

  test('GettersAndSettersDataSet', (){
    List <List <double>> values = [[2.1, 4.1, 2.3],[2.3, 4.3, 2.5],[2.2, 4.2, 2.4]];
    List <String> labels = ["At1","At2","At3"];

    DataSet dataSetTest = new DataSet.List("TEST",2,labels, values, 1);

    //TITLE
    expect(dataSetTest.title == "TEST", true);
    dataSetTest.title = "tittleTest";
    expect(dataSetTest.title, equals("tittleTest"));

    //isSUPERVISED
    dataSetTest.isSupervised = false;
    expect(dataSetTest.numValues == 3, true);
    expect(dataSetTest.numClassValues == null, true);
    expect(dataSetTest.instances[0].numClassValues == null, true);
    expect(dataSetTest.instances[0].classValues == null, true);
    expect(dataSetTest.instances[0].numValues == 3, true);
    expect(dataSetTest.instances[0].values, equals([2.1, 4.1, 2.3]));

    dataSetTest.isSupervised = true;
    expect(dataSetTest.numValues == 2, true);
    expect(dataSetTest.numClassValues == 1, true);
    expect(dataSetTest.instances[0].numClassValues == 1, true);
    expect(dataSetTest.instances[0].classValues, equals([2.3]));
    expect(dataSetTest.instances[0].numValues == 2, true);
    expect(dataSetTest.instances[0].values, equals([2.1, 4.1]));

    //INSTANCES

    dataSetTest.instances = [[2.2, 4.2, 2.4],[2.4, 4.4, 2.6],[2.3, 4.3, 2.5],[1.0,2.4,2.1]];
    expect(dataSetTest.instances[0].instanceValues,equals([2.2, 4.2, 2.4]));
    expect(dataSetTest.instances[1].instanceValues,equals([2.4, 4.4, 2.6]));
    expect(dataSetTest.instances[2].instanceValues,equals([2.3, 4.3, 2.5]));
    expect(dataSetTest.instances[3].instanceValues,equals([1.0,2.4,2.1]));

    //TRAIN AND TEST INSTANCES

    dataSetTest.trainTestSet = 2;
    expect(dataSetTest.trainSet.length == 2, true);
    expect(dataSetTest.testSet.length == 2, true);

    dataSetTest.trainTestSet = 1;
    expect(dataSetTest.trainSet.length == 1, true);
    expect(dataSetTest.testSet.length == 3, true);
  });

  test('Methods dataSet', () {
    List <List <double>> values = [[2.1, 4.1, 2.3], [2.3, 4.3, 2.5], [2.2, 4.2, 2.4]];
    List <String> labels = ["At1", "At2", "At3"];

    DataSet dataSetTest = new DataSet.List("TEST", 2, labels, values, 1);

    //ROWS
    expect(dataSetTest.getRow(0), equals([2.1,2.3,2.2]));
    dataSetTest.setRow(0,[0.1,0.3,0.2]);
    expect(dataSetTest.getRow(0), equals([0.1,0.3,0.2]));
    dataSetTest.setRow(0,[2.1,2.3,2.2]);

    dataSetTest.addRow([2.8,2.8,2.8],"NewRow",false);
    expect(dataSetTest.instances[0].instanceValues,equals([2.1, 4.1, 2.8, 2.3]));
    expect(dataSetTest.instances[1].instanceValues,equals([2.3, 4.3, 2.8, 2.5]));
    expect(dataSetTest.instances[2].instanceValues,equals([2.2, 4.2, 2.8, 2.4]));

    dataSetTest.addRow([2.8,2.8,2.8],"NewRow",true);
    expect(dataSetTest.instances[0].instanceValues,equals([2.1, 4.1,2.8, 2.3,2.8]));
    expect(dataSetTest.instances[1].instanceValues,equals([2.3, 4.3,2.8, 2.5,2.8]));
    expect(dataSetTest.instances[2].instanceValues,equals([2.2, 4.2,2.8, 2.4,2.8]));

    dataSetTest.removeRow(4);

    expect(dataSetTest.instances[0].instanceValues,equals([2.1, 4.1, 2.8, 2.3]));
    expect(dataSetTest.instances[1].instanceValues,equals([2.3, 4.3, 2.8, 2.5]));
    expect(dataSetTest.instances[2].instanceValues,equals([2.2, 4.2, 2.8, 2.4]));

    dataSetTest.removeRow(2);

    expect(dataSetTest.instances[0].instanceValues,equals([2.1, 4.1, 2.3]));
    expect(dataSetTest.instances[1].instanceValues,equals([2.3, 4.3, 2.5]));
    expect(dataSetTest.instances[2].instanceValues,equals([2.2, 4.2, 2.4]));

    dataSetTest.removeRow(2);

    expect(dataSetTest.instances[0].instanceValues,equals([2.1, 4.1]));
    expect(dataSetTest.instances[1].instanceValues,equals([2.3, 4.3]));
    expect(dataSetTest.instances[2].instanceValues,equals([2.2, 4.2]));
    expect(dataSetTest.isSupervised, false);

    dataSetTest.addRow([2.3,2.5,2.4],"NewOne",true);
    //INSTANCES
    dataSetTest.addInstance([0.1,0.2,0.1]);
    dataSetTest.addInstance([0.1,0.2,0.1]);
    expect(dataSetTest.instances[3].instanceValues, equals([0.1,0.2,0.1]));
    expect(dataSetTest.instances.length == 4, true);
    dataSetTest.removeInstance(0);
    expect(dataSetTest.instances.length == 3, true);
  });
}