

import '../Functions/activationfunction.dart';
import '../Functions/inputfunction.dart';
import '../LearningRule/perceptronlearning.dart';
import '../neuron.dart';
import '../layer.dart';
import '../network.dart';
import '../DataSet/dataset.dart';


class Perceptron {

  Network network;
  DataSet dataSet;
  List <Instance> train;
  List <Instance> test;

  List <double> errorPerStep;
  List <double> errorPerCycle;


  Perceptron(this.dataSet, double learningRate) {

    List <Neuron> inputNeurons = [];
    List <Neuron> outputNeurons = [];

    Neuron perceptronNeuron = new Neuron("PerceptronNeuron", inputFunction : new WeightCombination(), activationFunction: new Step());

    for(int i = 0; i < this.dataSet.numValues; i++) {
      //Create a new inputNeuron
      Neuron tempNeuron = new Neuron.inputNeuron("InputNeuron" + i.toString());

      //Connect input neuron with perceptron neuron
      Connection tempConnection = new Connection(tempNeuron, perceptronNeuron);
      tempNeuron.outputConnection.add(tempConnection);
      perceptronNeuron.inputConnection.add(tempConnection);

      //Add inputNeuron to the inputLayer
      inputNeurons.add(tempNeuron);
    }
    //add perceptron neuron to the outputLayer
    outputNeurons.add(perceptronNeuron);

    Layer inputLayer = new Layer("inputLayer", inputNeurons);
    Layer outputLayer = new Layer("outputLayer", outputNeurons);

    this.network = new Network();
    this.network.addLayer(inputLayer);
    this.network.addLayer(outputLayer);
    this.network.learningRule = new perceptronLearning(learningRate);

    this.dataSet.trainTestSet = 1;

    this.train = this.dataSet.trainSet;
    this.test = this.dataSet.testSet;
  }

  Learn() {
    int j = 0;
    do{
      LearnCycle();
      j++;
      print(j);
    }while(j < 8);

  }

  LearnCycle() {
    for(int i = 0; i < this.train.length; i++){
      LearnStep(i);
    }
  }

  LearnStep(int step){

    Instance tempStep = this.train[step];
    List <double> tempValue = tempStep.values;
    List <double> tempClassValues = tempStep.classValues;


    for(int i = 0; i < this.network.layers[0].neurons.length -1; i++){
      this.network.layers[0].neurons[i].input = tempValue[i];

    }

    this.network.calculateOutput();
    this.network.learningRule.updateWeights(this.network.layers[1].getNeuronByIndex(0),tempClassValues[0]);


  }

  Test() {

  }


}
