library ArcNeuralNetwork.Layer;

//Imports

import "neuron.dart";
import "network.dart";

/*
Class Layer: represent a layer of the neural network
      Parameters:
        - Id: identifiers the layer --> String
        - Neurons: neurons of the layer --> List<Neuron>
        - Network: network to which the layer belongs --> network
      Methods:
        - Add Neuron
        - Remove Neuron
        - Get Neuron
        - Set Neuron
        - Get Weight
        - Set Weight
        - Calculate Output: get output of all neurons in layer
*/

class Layer {

  String id;
  List<Neuron> neurons;

  //////////////////////////////////////////////
  ////////////// Constructors //////////////////
  //////////////////////////////////////////////

  //Create a simple layer

  Layer(this.id, [List<Neuron>neurons]) {
    if(neurons!=null){
      this.neurons = neurons;
    }else {
      neurons = [];
    }
  }

  ///////////////////////////////////  Add Neuron  /////////////////////////////////

  void addNeuron(Neuron neuron) {
    if (neuron == null) {
      throw ("Neuron is empty");
    }
    this.neurons.add(neuron);
  }

  ///////////////////////////////////  Remove Neuron  /////////////////////////////////

  void removeAllNeurons() {
    this.neurons = [];
  }

  void removeNeuronAt(int index) {
    this.neurons.removeAt(index);
  }

  ///////////////////////////////////  Get Neuron  /////////////////////////////////

  int getNeuronIndexOf(Neuron neuron) {
    return this.neurons.indexOf(neuron);
  }

  Neuron getNeuronByIndex(int index) {
    if (index > this.neurons.length) {
      throw ("Index out of bounds");
    }
    return this.neurons[index];
  }

  Neuron getNeuronById(String id) {
    for (var neuron in this.neurons) {
      if (neuron.id == id) {
        return neuron;
      }
    }
    return null;
  }

  ///////////////////////////////////  Set Neuron  /////////////////////////////////

  void setNeuronAt(int index, Neuron neuron) {
    if (neuron == null) {
      throw ("Neuron is empty");
    }
    if (index > this.neurons.length) {
      throw ("Index out of bounds");
    }
    this.neurons[index] = neuron;
  }

  ///////////////////////////////////  Check Neuron  /////////////////////////////////

  bool get hasNeurons {
    return this.neurons.isEmpty;
  }


  ///////////////////////////////////  Calculate output of all neurons of the layer  /////////////////////////////////

  void calculateOutput() {
    for (Neuron neuron in this.neurons) {
      neuron.calculateInput();
      neuron.calculateOutput();
    }
  }
}
